@@ -1,7 +1,10 @@
 # -*- coding:utf-8 -*-
 
+import sys
 import random
 from collections import defaultdict
+from functools import reduce, cmp_to_key
+from itertools import chain, tee
 
 from grapheekdb.lib.undef import UNDEFINED
 from grapheekdb.lib.readwrite import GraphReadWrite
@@ -11,7 +14,7 @@ from grapheekdb.lib.exceptions import GrapheekDataException, GrapheekDataPrepara
 from grapheekdb.lib.exceptions import GrapheekIndexCreationFailedException, GrapheekIndexRemovalFailedException, GrapheekNoSuchTraversalException
 from grapheekdb.lib.exceptions import GrapheekMissingKeyException, GrapheekInvalidDataTypeException
 from grapheekdb.lib.exceptions import GrapheekUnknownScriptException, GrapheekUnknownAlias
-from grapheekdb.lib.exceptions import GrapheekUnknownMethod
+from grapheekdb.lib.exceptions import GrapheekUnknownMethod, GrapheekMixedKindException
 from grapheekdb.lib.expressions import eval_expr
 
 from grapheekdb.backends.data.keys import build_key, PREPARED
@@ -22,10 +25,13 @@ from grapheekdb.backends.data.keys import DATA_SUFFIX, IN_EDGES_SUFFIX, IN_VERTI
 from grapheekdb.backends.data.keys import COUNT_SUFFIX, KIND_VERTEX, KIND_EDGE, KIND_INDEX
 from grapheekdb.backends.data.keys import METADATA_EDGE_ID_LIST_PREFIX, METADATA_VERTEX_ID_LIST_PREFIX, CHUNK_SIZE
 from grapheekdb.backends.data.filtertools import build_filter_funcs, filter_entities
+from grapheekdb.backends.data.ordertools import build_order_func
 from grapheekdb.backends.data.indexes import ExactIndex, normalize_value
 from grapheekdb.backends.data.operations import Addition, Removal
 from grapheekdb.backends.data.optimizer import Optimizer
 
+PYTHON3 = sys.version_info.major == 3
+
 
 # misc iterators :
 def _jump_iterator(_graph, _kind, _entity_id_iterator, _traversal, _neighbors_cache, **filters):
@@ -133,7 +139,7 @@ class DotExport(object):
                 label = self._node_label(node_id)
                 node_collections.append((node_id, label))
                 for edge_id in self._graph._neighbors(KIND_VERTEX, node_id, BOTH_EDGES_SUFFIX):
-                    if not edge_id in node_edge_ids:
+                    if edge_id not in node_edge_ids:
                         edge_label, edge_src_id, edge_tgt_id = self._edge_struct(edge_id)
                         if (edge_src_id in node_ids) and (edge_tgt_id in node_ids):
                             edge_collections.append((edge_label, edge_src_id, edge_tgt_id))
@@ -246,7 +252,7 @@ class EntityIterator(object):
         self._src_kind = src_kind
         self._filters = filters
         self._filter_funcs = build_filter_funcs(**filters)
-        self._id_iterator = id_iterator
+        self._initial_id_iterator, self._id_iterator = tee(id_iterator)
         self._parent = parent
         self._neighbors_cache = neighbors_cache if neighbors_cache is not None else {}
         self._release_neighbors_cache = neighbors_cache is None
@@ -256,6 +262,52 @@ class EntityIterator(object):
             self._context = context
         self._current_iteration = None
 
+    def __next__(self):
+        if self._current_iteration is None:
+            self._current_iteration = self._clone_initial()._iterate(on_item=True)
+        try:
+            return next(self._current_iteration)
+        except StopIteration:
+            self._current_iteration = None
+        raise StopIteration
+
+    def __iter__(self):
+        return self
+
+    def __contains__(self, entity):
+        try:
+            if entity._kind != self._src_kind:
+                return False
+        except:
+            return False
+        entity_id = entity.get_id()
+        for it_entity_id in self._clone_initial()._iterate(on_item=False):
+            if it_entity_id == entity_id:
+                return True
+        return False
+
+    def __le__(self, it2):
+        return self._graph.issubset(self, it2)
+
+    def __ge__(self, it2):
+        return self._graph.issuperset(self, it2)
+
+    def __or__(self, it2):
+        # | => union
+        return self._graph.union(self, it2)
+
+    def __and__(self, it2):
+        # & => intersection
+        return self._graph.intersection(self, it2)
+
+    def __sub__(self, it2):
+        # - => difference
+        return self._graph.difference(self, it2)
+
+    def __xor__(self, it2):
+        # ^ => symmetric_difference
+        return self._graph.symmetric_difference(self, it2)
+
     def _finalize(self):
         if self._release_neighbors_cache:
             # Implicit cache removal
@@ -266,8 +318,9 @@ class EntityIterator(object):
             yield item
         self._finalize()
 
-    def __iter__(self):
-        return self
+    def _clone_initial(self):
+        self._initial_id_iterator, id_iterator = tee(self._initial_id_iterator)
+        return EntityIterator(self._graph, self._src_kind, self._filters, id_iterator, self._parent)
 
     def _jump(self, _kind, _traversal, _random, *args, **filters):
         jumping_iterator = _random_iterator if _random else _jump_iterator
@@ -303,15 +356,6 @@ class EntityIterator(object):
         # Python 2 compatibility
         return self.__next__()
 
-    def __next__(self):
-        if self._current_iteration is None:
-            self._current_iteration = self._iterate(on_item=True)
-        try:
-            return next(self._current_iteration)
-        except StopIteration:
-            self._current_iteration = None
-        raise StopIteration
-
     def inV(self, *args, **filters):
         return self._jump(KIND_VERTEX, IN_VERTICES_SUFFIX, False, *args, **filters)
 
@@ -443,7 +487,8 @@ class EntityIterator(object):
 
     def without(self, *aliases):
         def without_iterator(entity_iterator, aliases):
-            for entity_id in entity_iterator._iterate():  # Iterating updates the context ...
+            clone_entity_iterator = entity_iterator
+            for entity_id in clone_entity_iterator._iterate(on_item=False):  # Iterating updates the context ...
                 member = False
                 for alias in aliases:
                     excluded_entity = entity_iterator._context.get(alias, UNDEFINED)
@@ -454,12 +499,12 @@ class EntityIterator(object):
                     yield entity_id
         return EntityIterator(self._graph, self._src_kind, {}, without_iterator(self, aliases), parent=self)
 
-    def idata(self):
+    def idata(self, *args, **kwargs):
         for item in self:
-            yield item.data()
+            yield item.data(*args, **kwargs)
 
-    def data(self):
-        return list(self.idata())
+    def data(self, *args, **kwargs):
+        return list(self.idata(*args, **kwargs))
 
     def limit(self, count):
         def limit_iterator(entity_id_iterator, count):
@@ -475,11 +520,8 @@ class EntityIterator(object):
     def aka(self, alias):
         def alias_iterator(entity_id_iterator, alias):
             graph = self._graph
-            entity_cache = {}
             for entity_id in entity_id_iterator:
-                entity = entity_cache.get(entity_id, None)
-                if entity is None:
-                    entity_cache[entity_id] = self._context[alias] = graph._item_from_id(self._src_kind, entity_id)
+                self._context[alias] = graph._item_from_id(self._src_kind, entity_id)
                 yield entity_id
         return EntityIterator(self._graph, self._src_kind, {}, alias_iterator(self._iterate(), alias), parent=self)
 
@@ -507,6 +549,9 @@ class EntityIterator(object):
     def all(self):
         return list(self)
 
+    def ids(self):
+        return list(self._iterate(on_item=False))
+
     def collect(self, *aliases):
         result = []
         for entity in self:  # Iterating updates the context ...
@@ -519,20 +564,34 @@ class EntityIterator(object):
     def percent(self, expr=None, asc=False):
         return EntityAggregate(self._graph, self._src_kind, self, pair_iterator=None, agg='%', expr=expr, asc=asc)
 
-    def call(self, _fname, *aliases, **params):
-        contexts = []
+    def order_by(self, *clauses):
+
+        def order_iterator(entity_id_iterator, cmp_func):
+            lst = []
+            for item in self:
+                dic = item.data().copy()
+                dic['_id'] = item.get_id()
+                lst.append(dic)
+            if PYTHON3:
+                lst.sort(key=cmp_to_key(cmp_func))
+            else:
+                lst.sort(cmp_func)
+            for dic in lst:
+                yield dic['_id']
+        cmp_func = build_order_func(*clauses)
+        return EntityIterator(self._graph, self._src_kind, {}, order_iterator(order_iterator, cmp_func), parent=self)
+
+    def _context_iterator(self):
         for entity_id in self._iterate():  # Iterating updates the context ...
-            contexts.append(self._context)
-        self._graph._call_server_script(contexts, _fname, *aliases, **params)
+            yield self._context.copy()
+
+    def call(self, _fname, *aliases, **params):
+        # Same as request but ignore return
+        self.request(_fname, *aliases, **params)
 
     def request(self, _fname, *aliases, **params):
-        contexts = []
-        # Same as call but returns results of each call
-        results = []
-        for entity_id in self._iterate():  # Iterating updates the context ...
-            contexts.append(self._context)
-        results.append(self._graph._call_server_script(contexts, _fname, *aliases, **params))
-        return results
+        fn = self._graph._get_server_script(_fname)
+        return fn(self._graph, self._context_iterator(), *aliases, **params)
 
     def dot(self, node_label=None, edge_label=None, limit=100):
         return DotExport(self, node_label, edge_label, limit)
@@ -628,8 +687,8 @@ class Entity(object):
             raise GrapheekUnknownMethod(_method)
         return self._jump(kind, traversal, True, **filters)
 
-    def data(self):
-        return self._graph._get_data(None, self._kind, self._entity_id)
+    def data(self, *args, **kwargs):
+        return self._graph._get_data(None, self._kind, self._entity_id, *args, **kwargs)
 
     def get_id(self):
         return self._entity_id
@@ -650,21 +709,21 @@ class Node(Entity):
         self._kind = KIND_VERTEX
 
     def __change_denorm_data(self, lst_method, counter_method, *counter_args):
-        node_id = self._entity_id
+        root_key = KIND_VERTEX + '/' + str(self._entity_id) + '/'
         # traversal denormalized lists
-        lst_method(build_key(KIND_VERTEX, node_id, IN_EDGES_SUFFIX))
-        lst_method(build_key(KIND_VERTEX, node_id, IN_VERTICES_SUFFIX))
-        lst_method(build_key(KIND_VERTEX, node_id, OUT_EDGES_SUFFIX))
-        lst_method(build_key(KIND_VERTEX, node_id, OUT_VERTICES_SUFFIX))
-        lst_method(build_key(KIND_VERTEX, node_id, BOTH_EDGES_SUFFIX))
-        lst_method(build_key(KIND_VERTEX, node_id, BOTH_VERTICES_SUFFIX))
+        lst_method(root_key + IN_EDGES_SUFFIX)
+        lst_method(root_key + IN_VERTICES_SUFFIX)
+        lst_method(root_key + OUT_EDGES_SUFFIX)
+        lst_method(root_key + OUT_VERTICES_SUFFIX)
+        lst_method(root_key + BOTH_EDGES_SUFFIX)
+        lst_method(root_key + BOTH_VERTICES_SUFFIX)
         # traversal denormalized counters
-        counter_method(build_key(KIND_VERTEX, node_id, IN_EDGES_SUFFIX, COUNT_SUFFIX), *counter_args)
-        counter_method(build_key(KIND_VERTEX, node_id, IN_VERTICES_SUFFIX, COUNT_SUFFIX), *counter_args)
-        counter_method(build_key(KIND_VERTEX, node_id, OUT_EDGES_SUFFIX, COUNT_SUFFIX), *counter_args)
-        counter_method(build_key(KIND_VERTEX, node_id, OUT_VERTICES_SUFFIX, COUNT_SUFFIX), *counter_args)
-        counter_method(build_key(KIND_VERTEX, node_id, BOTH_EDGES_SUFFIX, COUNT_SUFFIX), *counter_args)
-        counter_method(build_key(KIND_VERTEX, node_id, BOTH_VERTICES_SUFFIX, COUNT_SUFFIX), *counter_args)
+        counter_method(root_key + IN_EDGES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
+        counter_method(root_key + IN_VERTICES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
+        counter_method(root_key + OUT_EDGES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
+        counter_method(root_key + OUT_VERTICES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
+        counter_method(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
+        counter_method(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
 
     def _add_denorm_data(self):
         operation = Addition()
@@ -680,23 +739,26 @@ class Node(Entity):
         source_id = self._entity_id
         target_id = target.get_id()
         edge_id = edge.get_id()
+        root_key = KIND_VERTEX + '/' + str(source_id) + '/'
+        s_edge_id = str(edge_id)
+        s_target_id = str(target_id)
         # building operation :
         operation = Addition()
         # lists
-        operation.append_to_lst(build_key(KIND_VERTEX, source_id, OUT_EDGES_SUFFIX), edge_id)
-        operation.append_to_lst(build_key(KIND_VERTEX, source_id, BOTH_EDGES_SUFFIX), edge_id)
-        operation.append_to_lst(build_key(KIND_VERTEX, source_id, OUT_VERTICES_SUFFIX), target_id)
-        operation.append_to_lst(build_key(KIND_VERTEX, source_id, BOTH_VERTICES_SUFFIX), target_id)
+        operation.append_to_lst(root_key + OUT_EDGES_SUFFIX, edge_id)
+        operation.append_to_lst(root_key + BOTH_EDGES_SUFFIX, edge_id)
+        operation.append_to_lst(root_key + OUT_VERTICES_SUFFIX, target_id)
+        operation.append_to_lst(root_key + BOTH_VERTICES_SUFFIX, target_id)
         # counters
-        operation.update_inc(build_key(KIND_VERTEX, source_id, OUT_EDGES_SUFFIX, COUNT_SUFFIX))
-        operation.update_inc(build_key(KIND_VERTEX, source_id, BOTH_EDGES_SUFFIX, COUNT_SUFFIX))
-        operation.update_inc(build_key(KIND_VERTEX, source_id, OUT_VERTICES_SUFFIX, COUNT_SUFFIX))
-        operation.update_inc(build_key(KIND_VERTEX, source_id, BOTH_VERTICES_SUFFIX, COUNT_SUFFIX))
+        operation.update_inc(root_key + OUT_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_inc(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_inc(root_key + OUT_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_inc(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
         # create a key/value for links between source and target (aimed to be used with indexes)
-        operation.set(build_key(KIND_VERTEX, source_id, OUT_EDGES_SUFFIX, edge_id), 1)
-        operation.set(build_key(KIND_VERTEX, source_id, BOTH_EDGES_SUFFIX, edge_id), 1)
-        operation.set(build_key(KIND_VERTEX, source_id, OUT_VERTICES_SUFFIX, target_id), 1)
-        operation.set(build_key(KIND_VERTEX, source_id, BOTH_VERTICES_SUFFIX, target_id), 1)
+        operation.set(root_key + OUT_EDGES_SUFFIX + '/' + s_edge_id, 1)
+        operation.set(root_key + BOTH_EDGES_SUFFIX + '/' + s_edge_id, 1)
+        operation.set(root_key + OUT_VERTICES_SUFFIX + '/' + s_target_id, 1)
+        operation.set(root_key + BOTH_VERTICES_SUFFIX + '/' + s_target_id, 1)
         # ---
         return operation
 
@@ -704,23 +766,26 @@ class Node(Entity):
         source_id = self._entity_id
         target_id = target.get_id()
         edge_id = edge.get_id()
+        root_key = KIND_VERTEX + '/' + str(source_id) + '/'
+        s_edge_id = str(edge_id)
+        s_target_id = str(target_id)
         # building operation :
         operation = Removal()
         # lists
-        operation.remove_from_lst(build_key(KIND_VERTEX, source_id, OUT_EDGES_SUFFIX), edge_id)
-        operation.remove_from_lst(build_key(KIND_VERTEX, source_id, BOTH_EDGES_SUFFIX), edge_id)
-        operation.remove_from_lst(build_key(KIND_VERTEX, source_id, OUT_VERTICES_SUFFIX), target_id)
-        operation.remove_from_lst(build_key(KIND_VERTEX, source_id, BOTH_VERTICES_SUFFIX), target_id)
+        operation.remove_from_lst(root_key + OUT_EDGES_SUFFIX, edge_id)
+        operation.remove_from_lst(root_key + BOTH_EDGES_SUFFIX, edge_id)
+        operation.remove_from_lst(root_key + OUT_VERTICES_SUFFIX, target_id)
+        operation.remove_from_lst(root_key + BOTH_VERTICES_SUFFIX, target_id)
         # counters
-        operation.update_dec(build_key(KIND_VERTEX, source_id, OUT_EDGES_SUFFIX, COUNT_SUFFIX))
-        operation.update_dec(build_key(KIND_VERTEX, source_id, BOTH_EDGES_SUFFIX, COUNT_SUFFIX))
-        operation.update_dec(build_key(KIND_VERTEX, source_id, OUT_VERTICES_SUFFIX, COUNT_SUFFIX))
-        operation.update_dec(build_key(KIND_VERTEX, source_id, BOTH_VERTICES_SUFFIX, COUNT_SUFFIX))
+        operation.update_dec(root_key + OUT_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_dec(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_dec(root_key + OUT_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_dec(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
         # create a key/value for links between source and target (aimed to be used with indexes)
-        operation.remove(build_key(KIND_VERTEX, source_id, OUT_EDGES_SUFFIX, edge_id))
-        operation.remove(build_key(KIND_VERTEX, source_id, BOTH_EDGES_SUFFIX, edge_id))
-        operation.remove(build_key(KIND_VERTEX, source_id, OUT_VERTICES_SUFFIX, target_id))
-        operation.remove(build_key(KIND_VERTEX, source_id, BOTH_VERTICES_SUFFIX, target_id))
+        operation.remove(root_key + OUT_EDGES_SUFFIX + '/' + s_edge_id)
+        operation.remove(root_key + BOTH_EDGES_SUFFIX + '/' + s_edge_id)
+        operation.remove(root_key + OUT_VERTICES_SUFFIX + '/' + s_target_id)
+        operation.remove(root_key + BOTH_VERTICES_SUFFIX + '/' + s_target_id)
         # ---
         return operation
 
@@ -728,23 +793,26 @@ class Node(Entity):
         target_id = self._entity_id
         source_id = source.get_id()
         edge_id = edge.get_id()
+        root_key = KIND_VERTEX + '/' + str(target_id) + '/'
+        s_edge_id = str(edge_id)
+        s_source_id = str(source_id)
         # building operation :
         operation = Addition()
         # lists
-        operation.append_to_lst(build_key(KIND_VERTEX, target_id, IN_EDGES_SUFFIX), edge_id)
-        operation.append_to_lst(build_key(KIND_VERTEX, target_id, BOTH_EDGES_SUFFIX), edge_id)
-        operation.append_to_lst(build_key(KIND_VERTEX, target_id, IN_VERTICES_SUFFIX), source_id)
-        operation.append_to_lst(build_key(KIND_VERTEX, target_id, BOTH_VERTICES_SUFFIX), source_id)
+        operation.append_to_lst(root_key + IN_EDGES_SUFFIX, edge_id)
+        operation.append_to_lst(root_key + BOTH_EDGES_SUFFIX, edge_id)
+        operation.append_to_lst(root_key + IN_VERTICES_SUFFIX, source_id)
+        operation.append_to_lst(root_key + BOTH_VERTICES_SUFFIX, source_id)
         # counters
-        operation.update_inc(build_key(KIND_VERTEX, target_id, IN_EDGES_SUFFIX, COUNT_SUFFIX))
-        operation.update_inc(build_key(KIND_VERTEX, target_id, BOTH_EDGES_SUFFIX, COUNT_SUFFIX))
-        operation.update_inc(build_key(KIND_VERTEX, target_id, IN_VERTICES_SUFFIX, COUNT_SUFFIX))
-        operation.update_inc(build_key(KIND_VERTEX, target_id, BOTH_VERTICES_SUFFIX, COUNT_SUFFIX))
+        operation.update_inc(root_key + IN_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_inc(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_inc(root_key + IN_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_inc(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
         # create a key/value for links between source and target (aimed to be used with indexes)
-        operation.set(build_key(KIND_VERTEX, target_id, IN_EDGES_SUFFIX, edge_id), 1)
-        operation.set(build_key(KIND_VERTEX, target_id, BOTH_EDGES_SUFFIX, edge_id), 1)
-        operation.set(build_key(KIND_VERTEX, target_id, IN_VERTICES_SUFFIX, source_id), 1)
-        operation.set(build_key(KIND_VERTEX, target_id, BOTH_VERTICES_SUFFIX, source_id), 1)
+        operation.set(root_key + IN_EDGES_SUFFIX + '/' + s_edge_id, 1)
+        operation.set(root_key + BOTH_EDGES_SUFFIX + '/' + s_edge_id, 1)
+        operation.set(root_key + IN_VERTICES_SUFFIX + '/' + s_source_id, 1)
+        operation.set(root_key + BOTH_VERTICES_SUFFIX + '/' + s_source_id, 1)
         # ---
         return operation
 
@@ -752,23 +820,26 @@ class Node(Entity):
         target_id = self._entity_id
         source_id = source.get_id()
         edge_id = edge.get_id()
+        root_key = KIND_VERTEX + '/' + str(target_id) + '/'
+        s_edge_id = str(edge_id)
+        s_source_id = str(source_id)
         # building operation :
         operation = Removal()
         # lists
-        operation.remove_from_lst(build_key(KIND_VERTEX, target_id, IN_EDGES_SUFFIX), edge_id)
-        operation.remove_from_lst(build_key(KIND_VERTEX, target_id, BOTH_EDGES_SUFFIX), edge_id)
-        operation.remove_from_lst(build_key(KIND_VERTEX, target_id, IN_VERTICES_SUFFIX), source_id)
-        operation.remove_from_lst(build_key(KIND_VERTEX, target_id, BOTH_VERTICES_SUFFIX), source_id)
+        operation.remove_from_lst(root_key + IN_EDGES_SUFFIX, edge_id)
+        operation.remove_from_lst(root_key + BOTH_EDGES_SUFFIX, edge_id)
+        operation.remove_from_lst(root_key + IN_VERTICES_SUFFIX, source_id)
+        operation.remove_from_lst(root_key + BOTH_VERTICES_SUFFIX, source_id)
         # counters
-        operation.update_dec(build_key(KIND_VERTEX, target_id, IN_EDGES_SUFFIX, COUNT_SUFFIX))
-        operation.update_dec(build_key(KIND_VERTEX, target_id, BOTH_EDGES_SUFFIX, COUNT_SUFFIX))
-        operation.update_dec(build_key(KIND_VERTEX, target_id, IN_VERTICES_SUFFIX, COUNT_SUFFIX))
-        operation.update_dec(build_key(KIND_VERTEX, target_id, BOTH_VERTICES_SUFFIX, COUNT_SUFFIX))
+        operation.update_dec(root_key + IN_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_dec(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_dec(root_key + IN_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
+        operation.update_dec(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
         # create a key/value for links between source and target (aimed to be used with indexes)
-        operation.remove(build_key(KIND_VERTEX, target_id, IN_EDGES_SUFFIX, edge_id))
-        operation.remove(build_key(KIND_VERTEX, target_id, BOTH_EDGES_SUFFIX, edge_id))
-        operation.remove(build_key(KIND_VERTEX, target_id, IN_VERTICES_SUFFIX, source_id))
-        operation.remove(build_key(KIND_VERTEX, target_id, BOTH_VERTICES_SUFFIX, source_id))
+        operation.remove(root_key + IN_EDGES_SUFFIX + '/' + s_edge_id)
+        operation.remove(root_key + BOTH_EDGES_SUFFIX + '/' + s_edge_id)
+        operation.remove(root_key + IN_VERTICES_SUFFIX + '/' + s_source_id)
+        operation.remove(root_key + BOTH_VERTICES_SUFFIX + '/' + s_source_id)
         # ---
         return operation
 
@@ -833,18 +904,21 @@ class Edge(Entity):
         edge_id = self._entity_id
         source_id = source.get_id()
         target_id = target.get_id()
+        root_key = KIND_EDGE + '/' + str(edge_id) + '/'
+        s_source_id = str(source_id)
+        s_target_id = str(target_id)
         # building operation :
         operation = Addition()
         # traversal denormalized lists
-        operation.append_to_lst(build_key(KIND_EDGE, edge_id, IN_VERTICES_SUFFIX), source_id)
-        operation.append_to_lst(build_key(KIND_EDGE, edge_id, OUT_VERTICES_SUFFIX), target_id)
-        operation.append_to_lst(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX), source_id)
-        operation.append_to_lst(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX), target_id)
+        operation.append_to_lst(root_key + IN_VERTICES_SUFFIX, source_id)
+        operation.append_to_lst(root_key + OUT_VERTICES_SUFFIX, target_id)
+        operation.append_to_lst(root_key + BOTH_VERTICES_SUFFIX, source_id)
+        operation.append_to_lst(root_key + BOTH_VERTICES_SUFFIX, target_id)
         # create a key/value for links between source, target and current edge (aimed to be used with indexes)
-        operation.set(build_key(KIND_EDGE, edge_id, IN_VERTICES_SUFFIX, source_id), 1)
-        operation.set(build_key(KIND_EDGE, edge_id, OUT_VERTICES_SUFFIX, target_id), 1)
-        operation.set(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX, source_id), 1)
-        operation.set(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX, target_id), 1)
+        operation.set(root_key + IN_VERTICES_SUFFIX + '/' + s_source_id, 1)
+        operation.set(root_key + OUT_VERTICES_SUFFIX + '/' + s_target_id, 1)
+        operation.set(root_key + BOTH_VERTICES_SUFFIX + '/' + s_source_id, 1)
+        operation.set(root_key + BOTH_VERTICES_SUFFIX + '/' + s_target_id, 1)
         # ---
         return operation
 
@@ -852,18 +926,21 @@ class Edge(Entity):
         edge_id = self._entity_id
         source_id = source.get_id()
         target_id = target.get_id()
+        root_key = KIND_EDGE + '/' + str(edge_id) + '/'
+        s_source_id = str(source_id)
+        s_target_id = str(target_id)
         # building operation
         operation = Removal()
         # traversal denormalized lists
-        operation.remove(build_key(KIND_EDGE, edge_id, IN_VERTICES_SUFFIX))
-        operation.remove(build_key(KIND_EDGE, edge_id, OUT_VERTICES_SUFFIX))
-        operation.remove(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX))
-        operation.remove(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX))
+        operation.remove(root_key + IN_VERTICES_SUFFIX)
+        operation.remove(root_key + OUT_VERTICES_SUFFIX)
+        operation.remove(root_key + BOTH_VERTICES_SUFFIX)
+        operation.remove(root_key + BOTH_VERTICES_SUFFIX)
         # remove key/value for links between source, target and current edge (aimed to be used with indexes)
-        operation.remove(build_key(KIND_EDGE, edge_id, IN_VERTICES_SUFFIX, source_id))
-        operation.remove(build_key(KIND_EDGE, edge_id, OUT_VERTICES_SUFFIX, target_id))
-        operation.remove(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX, source_id))
-        operation.remove(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX, target_id))
+        operation.remove(root_key + IN_VERTICES_SUFFIX + '/' + s_source_id)
+        operation.remove(root_key + OUT_VERTICES_SUFFIX + '/' + s_target_id)
+        operation.remove(root_key + BOTH_VERTICES_SUFFIX + '/' + s_source_id)
+        operation.remove(root_key + BOTH_VERTICES_SUFFIX + '/' + s_target_id)
         # ---
         return operation
 
@@ -1034,10 +1111,18 @@ class BaseGraph(GraphReadWrite, GraphNx):
     # Sub data update (no need to override, IMHO)
     # Remark that we are using kind, id instead of key (key is a *key* of the Key Value Store, id is an entity id (node id or edge id for instance))
 
-    def _get_data(self, txn, kind, entity_id):
+    def _get_data(self, txn, kind, entity_id, *args, **kwargs):
         key = build_key(kind, entity_id, DATA_SUFFIX)
-        res = self._get(txn, key)  # if _get returns UNDEFINED, that's also ok, I return it...
-        return res
+        result = self._get(txn, key)  # if _get returns UNDEFINED, that's also ok, I return it...
+        if args:
+            tmpres = dict((k, v) for k, v in result.items() if k in args)
+            result = tmpres
+            if kwargs:
+                flat = kwargs.get('flat', False)
+                if flat:
+                    tmpres = [result[k] for k in args]
+                    result = tmpres
+        return result
 
     def _bulk_update_data(self, _txn, _kind, _entity_id, **updates):
         check_valid_data(updates)
@@ -1050,10 +1135,9 @@ class BaseGraph(GraphReadWrite, GraphNx):
             # Remove entity from indexes before updating data (we'll re-add it just after with new values)
             self._remove_from_all_entity_indexes(_txn, _kind, _entity_id)
             # Do the update :
-            for subkey, value in list(updates.items()):
-                key = build_key(_kind, _entity_id, DATA_SUFFIX)
-                data[subkey] = value
-                self._set(_txn, key, data)
+            data.update(**updates)
+            key = build_key(_kind, _entity_id, DATA_SUFFIX)
+            self._set(_txn, key, data)
             # Readding entity to indexes :
             self._add_to_all_entity_indexes(_txn, _kind, _entity_id, data)
             if release_txn:
@@ -1068,7 +1152,7 @@ class BaseGraph(GraphReadWrite, GraphNx):
 
     # DON'T OVERRIDE THOSE METHODS :
 
-    def _call_server_script(self, _ctx, _fname, *aliases, **params):
+    def _get_server_script(self, _fname):
         # 1st, get function from cache :
         func = self._script_cache.get(_fname, None)
         if func is None:
@@ -1084,8 +1168,7 @@ class BaseGraph(GraphReadWrite, GraphNx):
         if func is None:
             # not found -> raise an exception to warn user
             raise GrapheekUnknownScriptException('%s method cannot be found' % (_fname,))
-        # let's call function :
-        return func(self, _ctx, *aliases, **params)
+        return func
 
     def _add_node(self, txn, data, node_id=None):
         assert(isinstance(data, dict))
@@ -1098,7 +1181,7 @@ class BaseGraph(GraphReadWrite, GraphNx):
             if node_id is None:
                 node_id = self._new_id_for_key(txn, METADATA_VERTEX_COUNTER)
             node = Node(node_id, self)
-            operation.set(build_key(KIND_VERTEX, node_id, DATA_SUFFIX), data)
+            operation.set(build_key(KIND_VERTEX, node_id, DATA_SUFFIX), data.copy())
             operation.merge(node._add_denorm_data())
             # Adding node id to proper node ids list :
             operation.append_to_lst(build_key(METADATA_VERTEX_ID_LIST_PREFIX, node_id // CHUNK_SIZE), node_id)
@@ -1352,7 +1435,15 @@ class BaseGraph(GraphReadWrite, GraphNx):
                 # to challenge every indexed ids to check that it is among current entity neighbours
                 # so it will lead to
                 # which can cost a lot (<index_count> disk access for persistent backends)
-                iterator = self._optimizer.index_or_seq_scan_iterator(_kind, seq_count * 2, **filters)
+                _traversal_kind = {
+                    IN_VERTICES_SUFFIX: KIND_VERTEX,
+                    OUT_VERTICES_SUFFIX: KIND_VERTEX,
+                    BOTH_VERTICES_SUFFIX: KIND_VERTEX,
+                    IN_EDGES_SUFFIX: KIND_EDGE,
+                    OUT_EDGES_SUFFIX: KIND_EDGE,
+                    BOTH_EDGES_SUFFIX: KIND_EDGE,
+                }[_traversal]
+                iterator = self._optimizer.index_or_seq_scan_iterator(_traversal_kind, seq_count * 2, **filters)
             if iterator is None:
                 if not(_random):
                     for entity_id in iter(neighbour_ids):
@@ -1378,7 +1469,6 @@ class BaseGraph(GraphReadWrite, GraphNx):
             seq_count = self._node_count
         else:
             seq_count = self._edge_count
-
         iterator = self._optimizer.index_or_seq_scan_iterator(_kind, seq_count, **filters)
         if iterator is None:  # no index is better than sequential scan, so choose it
             iterator = self._optimizer.get_kind_ids(None, _kind)
@@ -1403,19 +1493,19 @@ class BaseGraph(GraphReadWrite, GraphNx):
 
     def add_node(self, **kwargs):
         check_valid_data(kwargs)
-        return self._add_node(None, kwargs)
+        return self._add_node(None, kwargs.copy())
 
     def bulk_add_node(self, node_defns, silent=False):
         for data in node_defns:
             check_valid_data(data)
-        return self._bulk_add_node(node_defns, silent)
+        node_defns_copy = [node_defn.copy() for node_defn in node_defns]
+        return self._bulk_add_node(node_defns_copy, silent)
 
     def add_edge(self, _source, _target, **kwargs):
         check_base_node(_source)
         check_base_node(_target)
         check_valid_data(kwargs)
-
-        return self._add_edge(None, _source, _target, kwargs)
+        return self._add_edge(None, _source, _target, kwargs.copy())
 
     def bulk_add_edge(self, edge_defns, silent=False):
         try:
@@ -1425,7 +1515,8 @@ class BaseGraph(GraphReadWrite, GraphNx):
                 check_valid_data(data)
         except (TypeError, ValueError) as e:
             raise GrapheekDataException(repr(e))
-        return self._bulk_add_edge(edge_defns, silent)
+        edge_defns_copy = [(source, target, data.copy()) for source, target, data in edge_defns]
+        return self._bulk_add_edge(edge_defns_copy, silent)
 
     def bulk_add_edge_by_ids(self, edge_id_defns, silent=False):
         return self._bulk_add_edge_by_ids(edge_id_defns, silent)
@@ -1528,11 +1619,13 @@ class BaseGraph(GraphReadWrite, GraphNx):
             METADATA_INDEX_PREFIX = METADATA_VERTEX_INDEX_PREFIX
             METADATA_INDEX_LIST = METADATA_VERTEX_INDEX_LIST
             METADATA_INDEX_FIELDS_PREFIX = METADATA_VERTEX_INDEX_FIELDS_PREFIX
+            METADATA_INDEX_COUNTER = METADATA_VERTEX_INDEX_COUNTER
         else:
             indexes = self._edge_indexes
             METADATA_INDEX_PREFIX = METADATA_EDGE_INDEX_PREFIX
             METADATA_INDEX_LIST = METADATA_EDGE_INDEX_LIST
             METADATA_INDEX_FIELDS_PREFIX = METADATA_EDGE_INDEX_FIELDS_PREFIX
+            METADATA_INDEX_COUNTER = METADATA_EDGE_INDEX_COUNTER
         index_to_remove = None
         for index in indexes:
             if index._fields == fields and index._filter_items == set(kwargs.items()):
@@ -1546,6 +1639,8 @@ class BaseGraph(GraphReadWrite, GraphNx):
             index.delete(txn)  # it is the responsability of the index to remove everything it created
             # Removing index from index list :
             indexes.remove(index)
+            # update counter
+            self._update_dec(txn, METADATA_INDEX_COUNTER)
             # Removing "management key"
             index_key = '/'.join([METADATA_INDEX_PREFIX] + [index_signature_string])
             index_id = self._get(txn, index_key)
@@ -1560,11 +1655,98 @@ class BaseGraph(GraphReadWrite, GraphNx):
             self._transaction_rollback(txn)
             raise GrapheekIndexRemovalFailedException(repr(e))
 
-    def remove_node_index(self, *args):
-        self._remove_entity_index(KIND_VERTEX, *args)
+    def remove_node_index(self, *fields, **filters):
+        self._remove_entity_index(KIND_VERTEX, *fields, **filters)
 
-    def remove_edge_index(self, *args):
-        self._remove_entity_index(KIND_EDGE, *args)
+    def remove_edge_index(self, *fields, **filters):
+        self._remove_entity_index(KIND_EDGE, *fields, **filters)
+
+    def _operator_kind_and_checks(self, *entity_iterators):
+        # ensure entity iterators are of the same kind :
+        kinds = set([it._src_kind for it in entity_iterators])
+        if len(set(kinds)) > 1:
+            raise GrapheekMixedKindException
+        kind = kinds.pop()
+        return kind
+
+    def union(self, *entity_iterators):
+        kind = self._operator_kind_and_checks(*entity_iterators)
+
+        def id_iterator(*entity_iterators):
+            known_ids = set()
+            clone_entity_iterators = [entity_iterator._clone_initial() for entity_iterator in entity_iterators]
+            for entity in chain(*clone_entity_iterators):
+                entity_id = entity._entity_id
+                if entity_id not in known_ids:
+                    known_ids.add(entity_id)
+                    yield entity_id
+        return EntityIterator(self, kind, {}, id_iterator(*entity_iterators))
+
+    def _marshalled_operator_helper(self, commands):  # pragma : no cover
+        entity_iterators = []
+        for command in commands:
+            obj = self
+            for method_name, args, kwargs in command:
+                method = getattr(obj, method_name)
+                obj = method(*args, **kwargs)
+            entity_iterators.append(obj)
+        return entity_iterators
+
+    def _marshalled_union(self, *commands, **kwargs):  # pragma : no cover
+        entity_iterators = self._marshalled_operator_helper(commands)
+        return self.union(*entity_iterators)
+
+    def _set_operation(self, func, *entity_iterators):
+        kind = self._operator_kind_and_checks(*entity_iterators)
+        id_sets = [set(it._clone_initial().ids()) for it in entity_iterators]
+        ids = reduce(func, id_sets)
+        return EntityIterator(self, kind, {}, iter(ids))
+
+    def issubset(self, *entity_iterators):
+        self._operator_kind_and_checks(*entity_iterators)
+        id_sets = [set(it._clone_initial().ids()) for it in entity_iterators]
+        func = lambda x, y: x.issubset(y)
+        value = reduce(func, id_sets)
+        return value
+
+    def _marshalled_issubset(self, *commands, **kwargs):  # pragma : no cover
+        entity_iterators = self._marshalled_operator_helper(commands)
+        return self.issubset(*entity_iterators)
+
+    def issuperset(self, *entity_iterators):
+        self._operator_kind_and_checks(*entity_iterators)
+        id_sets = [set(it._clone_initial().ids()) for it in entity_iterators]
+        func = lambda x, y: x.issuperset(y)
+        value = reduce(func, id_sets)
+        return value
+
+    def _marshalled_issuperset(self, *commands, **kwargs):  # pragma : no cover
+        entity_iterators = self._marshalled_operator_helper(commands)
+        return self.issuperset(*entity_iterators)
+
+    def intersection(self, *entity_iterators):
+        func = lambda x, y: x.intersection(y)
+        return self._set_operation(func, *entity_iterators)
+
+    def _marshalled_intersection(self, *commands, **kwargs):  # pragma : no cover
+        entity_iterators = self._marshalled_operator_helper(commands)
+        return self.intersection(*entity_iterators)
+
+    def difference(self, *entity_iterators):
+        func = lambda x, y: x.difference(y)
+        return self._set_operation(func, *entity_iterators)
+
+    def _marshalled_difference(self, *commands, **kwargs):  # pragma : no cover
+        entity_iterators = self._marshalled_operator_helper(commands)
+        return self.difference(*entity_iterators)
+
+    def symmetric_difference(self, *entity_iterators):
+        func = lambda x, y: x.symmetric_difference(y)
+        return self._set_operation(func, *entity_iterators)
+
+    def _marshalled_symmetric_difference(self, *commands, **kwargs):  # pragma : no cover
+        entity_iterators = self._marshalled_operator_helper(commands)
+        return self.symmetric_difference(*entity_iterators)
 
     def setup_server_scripts(self, *module_paths):
         import sys
