@@ -1,6 +1,7 @@
 # -*- coding:utf-8 -*-
 
 from functools import partial
+from itertools import tee
 
 from . import lookups
 
@@ -81,15 +82,16 @@ def entity_match(filter_funcs, data, internals=None):
 
 
 def filter_entities(graph, kind, iterator, filter_funcs, on_item=False):
+    _, cloned_iterator = tee(iterator)
     if filter_funcs:
         if on_item:
-            for entity_id in iterator:
+            for entity_id in cloned_iterator:
                 data = graph._get(None, kind + '/' + str(entity_id) + '/' + DATA_SUFFIX)
                 match = entity_match(filter_funcs, data, {'_id': entity_id})
                 if match:
                     yield graph._item_from_id(kind, entity_id)
         else:
-            for entity_id in iterator:
+            for entity_id in cloned_iterator:
                 data = graph._get(None, kind + '/' + str(entity_id) + '/' + DATA_SUFFIX)
                 match = entity_match(filter_funcs, data, {'_id': entity_id})
                 if match:
@@ -98,8 +100,8 @@ def filter_entities(graph, kind, iterator, filter_funcs, on_item=False):
         # I don't do : yield item if on_item else entity_id
         # Because item is not computed (when on_item is False, this is not useful to instantiate an Entity)
         if on_item:
-            for entity_id in iterator:
+            for entity_id in cloned_iterator:
                 yield graph._item_from_id(kind, entity_id)
         else:
-            for entity_id in iterator:
+            for entity_id in cloned_iterator:
                 yield entity_id
