@@ -38,19 +38,25 @@ class LmdbGraph(BaseGraph):
         txn.abort()
 
     def _has_key(self, key):
+        k = key
+        if PYTHON3:  # pragma : no cover
+            k = bytes(k, encoding='utf8')
         with self._env.begin() as txn:
-            return txn.get(key, UNDEFINED) != UNDEFINED
+            return txn.get(k, UNDEFINED) != UNDEFINED
 
     def _get(self, txn, key):
+        k = key
+        if PYTHON3:  # pragma : no cover
+            k = bytes(k, encoding='utf8')
         if txn is not None:
-            raw_data = txn.get(key, UNDEFINED)
+            raw_data = txn.get(k, UNDEFINED)
             if raw_data == UNDEFINED:
                 return UNDEFINED
             res = msgpack.loads(raw_data, encoding='utf8')
             return res
         else:
             with self._env.begin() as txn:
-                raw_data = txn.get(key, UNDEFINED)
+                raw_data = txn.get(k, UNDEFINED)
                 if raw_data == UNDEFINED:
                     return UNDEFINED
                 res = msgpack.loads(raw_data, encoding='utf8')
@@ -65,18 +71,24 @@ class LmdbGraph(BaseGraph):
         return dic
 
     def _set(self, txn, key, value):
-        txn.put(key, msgpack.dumps(value, encoding='utf8'))
+        k = key
+        if PYTHON3:  # pragma : no cover
+            k = bytes(k, encoding='utf8')
+        txn.put(k, msgpack.dumps(value, encoding='utf8'))
 
     def _bulk_set(self, txn, updates):
         for key, value in list(updates.items()):
             self._set(txn, key, value)
 
     def _remove(self, txn, key):
-        txn.delete(key)
+        k = key
+        if PYTHON3:  # pragma : no cover
+            k = bytes(k, encoding='utf8')
+        txn.delete(k)
 
     def _bulk_remove(self, txn, keys):
         for key in keys:
-            txn.delete(key)
+            self._remove(txn, key)
 
     def _remove_prefix(self, txn, prefix):
         cursor = txn.cursor()
