@@ -95,6 +95,28 @@ class ProxyIteratorCommon(object):
     def __iter__(self):
         return self
 
+    def __le__(self, it2):
+        return self._graph.issubset(self, it2)
+
+    def __ge__(self, it2):
+        return self._graph.issuperset(self, it2)
+
+    def __or__(self, it2):
+        # | => union
+        return self._graph.union(self, it2)
+
+    def __and__(self, it2):
+        # & => intersection
+        return self._graph.intersection(self, it2)
+
+    def __sub__(self, it2):
+        # - => difference
+        return self._graph.difference(self, it2)
+
+    def __xor__(self, it2):
+        # ^ => symmetric_difference
+        return self._graph.symmetric_difference(self, it2)
+
     def next(self):
         # Python 2 compatibility
         return self.__next__()
@@ -183,8 +205,8 @@ class ProxyEntityIterator(ProxyIteratorCommon):
     def without(self, *aliases):
         return self._do('without', *aliases)
 
-    def data(self):
-        return self._raw('data')
+    def data(self, *args, **kwargs):
+        return self._raw('data', *args, **kwargs)
 
     def limit(self, count):
         return self._do('limit', count)
@@ -204,6 +226,9 @@ class ProxyEntityIterator(ProxyIteratorCommon):
     def all(self):
         return self._raw('all')
 
+    def ids(self):
+        return self._raw('ids')
+
     def collect(self, *aliases):
         return self._raw('collect', *aliases)
 
@@ -215,6 +240,9 @@ class ProxyEntityIterator(ProxyIteratorCommon):
         command = ('percent', [expr], dict(asc=asc))
         return ProxyEntityAggregate(self._graph, self._commands + [command])
 
+    def order_by(self, *clauses):
+        return self._do('order_by', *clauses)
+
     def call(self, _fname, *aliases, **params):
         return self._raw('call', _fname, *aliases, **params)
 
@@ -301,8 +329,8 @@ class ProxyEntity(object):
     def random(self, *args, **kwargs):
         return self._do('random', *args, **kwargs)
 
-    def data(self):
-        return self._data or self._raw('data')
+    def data(self, *args, **kwargs):
+        return self._raw('data', *args, **kwargs)
 
     def get_id(self):
         return int(self._entity_id)
@@ -346,14 +374,17 @@ class ProxyGraph(GraphReadWrite, GraphNx):
 
     def __init__(self, address):  # pragma : no cover
         self._address = address
-        self._zmq_context = zmq.Context()
-        self._zmq_socket = self._zmq_context.socket(zmq.REQ)
-        self._zmq_socket.connect(address)
+        #self._zmq_context = zmq.Context()
+        #self._zmq_socket = self._zmq_context.socket(zmq.REQ)
+        #self._zmq_socket.connect(address)
 
     def _request(self, commands):
         data = msgpack.dumps(sanitize(commands), encoding='utf8')
-        self._zmq_socket.send(data)
-        raw = self._zmq_socket.recv()
+        zmq_context = zmq.Context()
+        zmq_socket = zmq_context.socket(zmq.REQ)
+        zmq_socket.connect(self._address)
+        zmq_socket.send(data)
+        raw = zmq_socket.recv()
         msg = msgpack.loads(raw, encoding='utf8')
         return unmarshall(self, msg)
 
@@ -406,9 +437,9 @@ class ProxyGraph(GraphReadWrite, GraphNx):
         command = ['bulk_add_edge_by_ids', args, {'silent': silent}]
         return self._request([command])
 
-    def _add_entity_index(self, _kind, *args, **filters):
+    def _add_entity_index(self, _kind, *fields, **filters):
         assert(_kind in (KIND_EDGE, KIND_VERTEX))
-        command = ['add_node_index' if _kind == KIND_VERTEX else 'add_edge_index', args, filters]
+        command = ['add_node_index' if _kind == KIND_VERTEX else 'add_edge_index', fields, filters]
         return self._request([command])
 
     def add_node_index(self, *fields, **filters):
@@ -425,13 +456,37 @@ class ProxyGraph(GraphReadWrite, GraphNx):
         command = ['get_edge_indexes', [], {}]
         return self._request([command])
 
-    def _remove_entity_index(self, _kind, *args):
+    def _remove_entity_index(self, _kind, *fields, **filters):
         assert(_kind in (KIND_EDGE, KIND_VERTEX))
-        command = ['remove_node_index' if _kind == KIND_VERTEX else 'remove_edge_index', args, {}]
+        command = ['remove_node_index' if _kind == KIND_VERTEX else 'remove_edge_index', fields, filters]
         return self._request([command])
 
-    def remove_node_index(self, *args):
-        return self._remove_entity_index(KIND_VERTEX, *args)
+    def remove_node_index(self, *fields, **filters):
+        return self._remove_entity_index(KIND_VERTEX, *fields, **filters)
 
-    def remove_edge_index(self, *args):
-        return self._remove_entity_index(KIND_EDGE, *args)
+    def remove_edge_index(self, *fields, **filters):
+        return self._remove_entity_index(KIND_EDGE, *fields, **filters)
+
+    def _operation_helper(self, operation, *entity_iterators):
+        command = [operation, [entity_iterator._commands for entity_iterator in entity_iterators], {}]
+        return ProxyEntityIterator(self, [command])
+
+    def issubset(self, *entity_iterators):
+        command = ['_marshalled_issubset', [entity_iterator._commands for entity_iterator in entity_iterators], {}]
+        return self._request([command])
+
+    def issuperset(self, *entity_iterators):
+        command = ['_marshalled_issuperset', [entity_iterator._commands for entity_iterator in entity_iterators], {}]
+        return self._request([command])
+
+    def union(self, *entity_iterators):
+        return self._operation_helper('_marshalled_union', *entity_iterators)
+
+    def intersection(self, *entity_iterators):
+        return self._operation_helper('_marshalled_intersection', *entity_iterators)
+
+    def difference(self, *entity_iterators):
+        return self._operation_helper('_marshalled_difference', *entity_iterators)
+
+    def symmetric_difference(self, *entity_iterators):
+        return self._operation_helper('_marshalled_symmetric_difference', *entity_iterators)
