@@ -5,5 +5,5 @@ It exposes a single function that echoes arguments and keyword arguments
 """
 
 
-def echo(g, _ctx, *args, **kwargs):
-    return dict(args=args, kwargs=kwargs)
+def echo(g, _ctxs, *args, **kwargs):
+    return [dict(args=args, kwargs=kwargs) for ctx in _ctxs]
