@@ -171,7 +171,6 @@ def main(cmds=None, verbose=True, context=None):
         '-a',
         '--address',
         dest='address',
-        default=default_address,
         required=False,
         help="""The address string. This has the form 'protocol://interface:port',
         for example 'tcp://127.0.0.1:5555'. Protocols supported include
@@ -182,7 +181,6 @@ def main(cmds=None, verbose=True, context=None):
         "-b",
         "--backend",
         dest="backend",
-        default=default_backend,
         required=False,
         help='Data backend. must be of the form (for instance) : "grapheekdb.backends.data.kyotocab.KyotoCabinetGraph"',
     )
